@@ -28,20 +28,11 @@
 #include <stdio.h>
 #include <string.h>
 
-#include <shlobj.h>
 #include <windows.h>
+#include <shlobj.h>
 #include <limits.h>
-#include <direct.h>
 
 
-int win32_mkdir(const char *dir)
-{
-    wchar_t wdir[MAX_PATH];
-
-    MultiByteToWideChar(CP_UTF8, 0, dir, -1, wdir, MAX_PATH);
-    return _wmkdir(wdir);
-}
-
 char *file_get_config_home(void)
 {
     return file_get_data_home();
@@ -56,7 +47,9 @@ char *file_get_data_home(void)
                                  NULL, SHGFP_TYPE_CURRENT, wdir)) {
         int len = WideCharToMultiByte (CP_UTF8, 0, wdir, -1, NULL, 0, NULL, NULL);
         char *appdir = malloc(len);
-        WideCharToMultiByte (CP_UTF8, 0, wdir, -1, appdir, len, NULL, NULL);
+        if (appdir) {
+            WideCharToMultiByte (CP_UTF8, 0, wdir, -1, appdir, len, NULL, NULL);
+        }
         return appdir;
     }
 
@@ -85,7 +78,9 @@ const char *file_get_config_system(const char *dir)
                     NULL, SHGFP_TYPE_CURRENT, wdir)) {
             int len = WideCharToMultiByte (CP_UTF8, 0, wdir, -1, NULL, 0, NULL, NULL);
             appdir = malloc(len);
-            WideCharToMultiByte (CP_UTF8, 0, wdir, -1, appdir, len, NULL, NULL);
+            if (appdir) {
+                WideCharToMultiByte (CP_UTF8, 0, wdir, -1, appdir, len, NULL, NULL);
+            }
             return appdir;
         } else {
             BD_DEBUG(DBG_FILE, "Can't find common configuration directory !\n");
