@@ -31,15 +31,24 @@
 #include <stdlib.h>
 #include <string.h>
 
-#include <fcntl.h>
 #include <unistd.h>
-#include <sys/stat.h>
 #include <sys/types.h>
+#include <sys/stat.h>
+#include <fcntl.h>
+
+#ifdef __ANDROID__
+# undef  lseek
+# define lseek lseek64
+# undef  off_t
+# define off_t off64_t
+#endif
 
 static void _file_close(AACS_FILE_H *file)
 {
     if (file) {
-        close((int)(intptr_t)file->internal);
+        if (close((int)(intptr_t)file->internal)) {
+            BD_DEBUG(DBG_CRIT | DBG_FILE, "Error closing POSIX file (%p)\n", (void*)file);
+        }
 
         BD_DEBUG(DBG_FILE, "Closed POSIX file (%p)\n", (void*)file);
 
@@ -67,7 +76,7 @@ static int64_t _file_read(AACS_FILE_H *file, uint8_t *buf, int64_t size)
     ssize_t got, result;
 
     if (size <= 0 || size >= BD_MAX_SSIZE) {
-        BD_DEBUG(DBG_FILE | DBG_CRIT, "Ignoring invalid read of size %"PRId64" (%p)\n", size, (void*)file);
+        BD_DEBUG(DBG_FILE | DBG_CRIT, "Ignoring invalid read of size %" PRId64 " (%p)\n", size, (void*)file);
         return 0;
     }
 
@@ -92,7 +101,14 @@ static int64_t _file_write(AACS_FILE_H *file, const uint8_t *buf, int64_t size)
     ssize_t written, result;
 
     if (size <= 0 || size >= BD_MAX_SSIZE) {
-        BD_DEBUG(DBG_FILE | DBG_CRIT, "Ignoring invalid write of size %"PRId64" (%p)\n", size, (void*)file);
+        if (size == 0) {
+            if (fsync((int)(intptr_t)file->internal)) {
+                BD_DEBUG(DBG_FILE, "fsync() failed (%p)\n", (void*)file);
+                return -1;
+            }
+            return 0;
+        }
+        BD_DEBUG(DBG_FILE | DBG_CRIT, "Ignoring invalid write of size %" PRId64 " (%p)\n", size, (void*)file);
         return 0;
     }
 
