@@ -30,7 +30,7 @@ typedef struct digit_key_pair_t digit_key_pair;
 struct digit_key_pair_t
 {
   unsigned int digit;
-  char *key;
+  uint8_t key[16];
 };
 
 /* list of digit_key_pair struct used in title entry */
@@ -89,8 +89,8 @@ struct title_entry_t
   uint8_t discid[20];
   //char *title;
   //date_entry date;
-  char *mek;
-  char *vid;
+  uint8_t mk[16];
+  uint8_t vid[16];
   //digit_key_pair_list *bn;
   uint8_t vuk[16];
   //digit_key_pair_list *pak;
@@ -116,14 +116,13 @@ struct config_file_t
   title_entry_list *list;
 };
 
-/* Functions used throughout the parser */
-BD_PRIVATE int keydbcfg_parse_config(config_file *cfgfile, const char *path);
+BD_PRIVATE int keydbcfg_parse_config(config_file *cfgfile, const char *path, const uint8_t *disc_id, int all_discs);
 BD_PRIVATE config_file *keydbcfg_new_config_file(void);
 BD_PRIVATE int keydbcfg_config_file_close(config_file *cfgfile);
 
 /* */
 
-BD_PRIVATE config_file *keydbcfg_config_load(const char *configfile_path);
+BD_PRIVATE config_file *keydbcfg_config_load(const char *configfile_path, const uint8_t *disc_id);
 
 BD_PRIVATE int   keycache_save(const char *type, const uint8_t *disc_id,
                                  const uint8_t *key, unsigned int len);
