@@ -42,10 +42,6 @@
 #include <inttypes.h>
 #include <string.h>
 #include <stdio.h>
-#ifdef HAVE_SYS_SELECT_H
-#include <sys/select.h>
-#endif
-#include <gcrypt.h>
 
 
 #define SECTOR_LEN       2048  /* bus encryption block size */
@@ -79,8 +75,9 @@ struct aacs {
 
     /* bus encryption */
     int       bee;        /* bus encryption enabled flag in content certificate */
-    int       bec;        /* bus encryption capable flag in drive certificate */
+    int       bec;        /* bus encryption capable flag in drive certificate. -1 = unread. */
     uint8_t   read_data_key[16];
+    uint8_t   drive_cert_hash[20];
 
     /* content certificate */
     CONTENT_CERT *cc;
@@ -93,8 +90,6 @@ struct aacs {
 static const uint8_t empty_key[20] = { 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
                                        0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
                                        0x00, 0x00, 0x00, 0x00 };
-static const uint8_t aacs_iv[16]   = { 0x0b, 0xa0, 0xf8, 0xdd, 0xfe, 0xa6, 0x1f, 0xb3,
-                                       0xd8, 0xdf, 0x9f, 0x56, 0x6a, 0x05, 0x0f, 0x78 };
 
 /*
  * Validate processing key using media key verification data
@@ -104,8 +99,7 @@ static int _validate_pk(const uint8_t *pk,
                         const uint8_t *cvalue, const uint8_t *uv, const uint8_t *vd,
                         uint8_t *mk)
 {
-    gcry_cipher_hd_t gcry_h;
-    int a;
+    int a, crypto_err;
     uint8_t dec_vd[16];
     char str[40];
 
@@ -115,18 +109,21 @@ static int _validate_pk(const uint8_t *pk,
     BD_DEBUG(DBG_AACS, "   cvalue: %s\n", str_print_hex(str, cvalue, 16));
     BD_DEBUG(DBG_AACS, "   Verification data: %s\n", str_print_hex(str, vd, 16));
 
-    gcry_cipher_open(&gcry_h, GCRY_CIPHER_AES, GCRY_CIPHER_MODE_ECB, 0);
-    gcry_cipher_setkey(gcry_h, pk, 16);
-    gcry_cipher_decrypt(gcry_h, mk, 16, cvalue, 16);
+    crypto_err = crypto_aes128d(pk, cvalue, mk);
+    if (crypto_err) {
+        LOG_CRYPTO_ERROR(DBG_AACS, "decrypting media key failed", crypto_err);
+        return AACS_ERROR_UNKNOWN;
+    }
 
     for (a = 0; a < 4; a++) {
         mk[a + 12] ^= uv[a];
     }
 
-    gcry_cipher_setkey(gcry_h, mk, 16);
-    gcry_cipher_decrypt (gcry_h, dec_vd, 16, vd, 16);
-    gcry_cipher_close(gcry_h);
-
+    crypto_err = crypto_aes128d(mk, vd, dec_vd);
+    if (crypto_err) {
+        LOG_CRYPTO_ERROR(DBG_AACS, "decrypting media key verification data failed", crypto_err);
+        return AACS_ERROR_UNKNOWN;
+    }
     if (!memcmp(dec_vd, "\x01\x23\x45\x67\x89\xAB\xCD\xEF", 8)) {
         BD_DEBUG(DBG_AACS, "Processing key %s is valid!\n", str_print_hex(str, pk, 16));
         return AACS_SUCCESS;
@@ -141,6 +138,8 @@ static int _validate_pk(const uint8_t *pk,
 
 static int _rl_verify_signature(const uint8_t *rl, size_t size)
 {
+    int crypto_err;
+
     if (size < 40) {
         BD_DEBUG(DBG_AACS, "too small revocation list\n");
         return 0;
@@ -158,7 +157,13 @@ static int _rl_verify_signature(const uint8_t *rl, size_t size)
         return 0;
     }
 
-    return crypto_aacs_verify_aacsla(rl + len, rl, len);
+    crypto_err = crypto_aacs_verify_aacsla(rl + len, rl, len);
+    if (crypto_err) {
+        LOG_CRYPTO_ERROR(DBG_AACS, "revocation list signature verification failed", crypto_err);
+        return 0;
+    }
+
+    return 1;
 }
 
 static void _save_rl(const char *name, uint32_t version, const uint8_t *version_rec,
@@ -178,9 +183,7 @@ static void _save_rl(const char *name, uint32_t version, const uint8_t *version_
     if (data) {
         memcpy(data,      version_rec, 12);
         memcpy(data + 12, rl_rec,      rl_len);
-        if (!_rl_verify_signature(data, rl_len + 12)) {
-            BD_DEBUG(DBG_AACS | DBG_CRIT, "invalid %s signature, not using it\n", name);
-        } else {
+        if (_rl_verify_signature(data, rl_len + 12)) {
             cache_save(name, version, data, rl_len + 12);
         }
         X_FREE(data);
@@ -224,11 +227,16 @@ static uint32_t _calc_v_mask(uint32_t uv)
     return v_mask;
 }
 
-static void _calc_pk(const uint8_t *dk, uint8_t *pk, uint32_t uv, uint32_t v_mask, uint32_t dev_key_v_mask)
+static int _calc_pk(const uint8_t *dk, uint8_t *pk, uint32_t uv, uint32_t v_mask, uint32_t dev_key_v_mask)
 {
     unsigned char left_child[16], right_child[16];
+    int crypto_err;
 
-    crypto_aesg3(dk, left_child, right_child, pk);
+    crypto_err = crypto_aesg3(dk, left_child, right_child, pk);
+    if (crypto_err) {
+        LOG_CRYPTO_ERROR(DBG_AACS, "PK derivation failed", crypto_err);
+        return AACS_ERROR_UNKNOWN;
+    }
 
     while (dev_key_v_mask != v_mask) {
 
@@ -246,13 +254,18 @@ static void _calc_pk(const uint8_t *dk, uint8_t *pk, uint32_t uv, uint32_t v_mas
             memcpy(curr_key, right_child, 16);
         }
 
-        crypto_aesg3(curr_key, left_child, right_child, pk);
+        crypto_err = crypto_aesg3(curr_key, left_child, right_child, pk);
+        if (crypto_err) {
+            LOG_CRYPTO_ERROR(DBG_AACS, "PK derivation failed", crypto_err);
+            return AACS_ERROR_UNKNOWN;
+        }
 
         dev_key_v_mask = ((int) dev_key_v_mask) >> 1;
     }
 
     char str[40];
     BD_DEBUG(DBG_AACS, "Processing key: %s\n",  str_print_hex(str, pk, 16));
+    return AACS_SUCCESS;
 }
 
 static dk_list *_find_dk(dk_list *dkl, uint32_t *p_dev_key_v_mask, uint32_t uv, uint32_t u_mask)
@@ -386,7 +399,10 @@ static int _calc_mk_dks(MKB *mkb, dk_list *dkl, uint8_t *mk)
         /* calculate processing key */
 
         uint8_t pk[16];
-        _calc_pk(dk->key, pk, uv, v_mask, dev_key_v_mask);
+        if (_calc_pk(dk->key, pk, uv, v_mask, dev_key_v_mask) != AACS_SUCCESS) {
+            /* try next device */
+            continue;
+        }
 
         /* calculate and verify media key */
 
@@ -472,17 +488,79 @@ static size_t _read_file(AACS *aacs, const char *file, void **data)
     return *data ? size : 0;
 }
 
+static size_t _read_mkb_file(AACS *aacs, const char *file, void **pdata)
+{
+    AACS_FILE_H *fp;
+    size_t       size = 0;
+    size_t       data_size = 65536; /* initial alloc */
+    uint32_t     chunk_size = 4; /* initial read */
+    uint8_t     *data;
+    int64_t      fsize;
+
+    *pdata = NULL;
+
+    fp = _file_open(aacs, file);
+    if (!fp) {
+        BD_DEBUG(DBG_AACS | DBG_CRIT, "Unable to open %s\n", file);
+        return 0;
+    }
+
+    fsize = file_size(fp);
+    if (fsize < 4) {
+        BD_DEBUG(DBG_AACS | DBG_CRIT, "Empty file: %s\n", file);
+        file_close(fp);
+        return 0;
+    }
+
+    data = malloc(data_size);
+    if (!data) {
+        BD_DEBUG(DBG_AACS | DBG_CRIT, "Out of memory\n");
+        file_close(fp);
+        return 0;
+    }
+
+    do {
+        int64_t read_size = chunk_size;
+        if (file_read(fp, data + size, read_size) != read_size) {
+            BD_DEBUG(DBG_AACS | DBG_CRIT, "Failed reading %s\n", file);
+            X_FREE(data);
+            break;
+        }
+        size += read_size;
+        chunk_size = MKINT_BE24(data + size - 4 + 1);
+        if (fsize - size + 4 < (int64_t)chunk_size) {
+            BD_DEBUG(DBG_AACS | DBG_CRIT, "Invalid record size %u in %s\n", (unsigned)chunk_size, file);
+            break;
+        }
+        if (data_size < size + chunk_size) {
+            data_size = 2*size + chunk_size;
+            void *tmp = realloc(data, data_size);
+            if (!tmp) {
+                X_FREE(data);
+                break;
+            }
+            data = tmp;
+        }
+    } while (chunk_size >= 4);
+
+    file_close(fp);
+
+    *pdata = data;
+
+    return data ? size : 0;
+}
+
 static MKB *_mkb_open(AACS *aacs)
 {
     size_t  size;
     void   *data;
     MKB    *mkb;
 
-    size = _read_file(aacs, "AACS" DIR_SEP "MKB_RO.inf", &data);
+    size = _read_mkb_file(aacs, "AACS" DIR_SEP "MKB_RO.inf", &data);
     if (size < 4) {
         /* retry with backup file */
         X_FREE(data);
-        size = _read_file(aacs, "AACS" DIR_SEP "DUPLICATE" DIR_SEP "MKB_RO.inf", &data);
+        size = _read_mkb_file(aacs, "AACS" DIR_SEP "DUPLICATE" DIR_SEP "MKB_RO.inf", &data);
     }
     if (size < 4) {
         X_FREE(data);
@@ -561,6 +639,12 @@ static int _calc_mk(AACS *aacs, uint8_t *mk, pk_list *pkl, dk_list *dkl)
         return AACS_SUCCESS;
     }
 
+    /* get cached mk */
+    if (!aacs->no_cache && keycache_find("mk", aacs->disc_id, mk, 16)) {
+        BD_DEBUG(DBG_AACS, "Using cached MK\n");
+        return AACS_SUCCESS;
+    }
+
     BD_DEBUG(DBG_AACS, "Calculate media key...\n");
 
     mkb = _mkb_open(aacs);
@@ -585,6 +669,12 @@ static int _calc_mk(AACS *aacs, uint8_t *mk, pk_list *pkl, dk_list *dkl)
 
     if (result == AACS_SUCCESS) {
         memcpy(aacs->mk, mk, sizeof(aacs->mk));
+        /* cache mk */
+        if (!aacs->no_cache) {
+            if (memcmp(aacs->disc_id, empty_key, sizeof(aacs->disc_id))) {
+                keycache_save("mk", aacs->disc_id, mk, 16);
+            }
+        }
     }
 
     mkb_close(mkb);
@@ -609,8 +699,11 @@ static int _mmc_read_auth(AACS *aacs, cert_list *hcl, int type, uint8_t *p1, uin
     for (; hcl ; hcl = hcl->next) {
 
         char tmp_str[2*92+1];
+        int crypto_error;
 
-        if (!crypto_aacs_verify_host_cert(hcl->host_cert)) {
+        crypto_error = crypto_aacs_verify_host_cert(hcl->host_cert);
+        if (crypto_error) {
+            LOG_CRYPTO_ERROR(DBG_AACS, "host certificate signature verification failed", crypto_error);
             BD_DEBUG(DBG_AACS, "Not using invalid host certificate %s.\n",
                   str_print_hex(tmp_str, hcl->host_cert, 92));
             continue;
@@ -668,10 +761,33 @@ static int _read_vid(AACS *aacs, cert_list *hcl)
 
 static int _read_read_data_key(AACS *aacs, cert_list *hcl)
 {
-    int error_code = _mmc_read_auth(aacs, hcl, MMC_READ_DATA_KEYS, aacs->read_data_key, NULL);
+    int error_code;
+    char *cache_path = NULL;
+    char str[41];
+
+    cache_path = str_printf("rdk" DIR_SEP "%s", str_print_hex(str, aacs->drive_cert_hash, 20));
+
+    error_code =_mmc_read_auth(aacs, hcl, MMC_READ_DATA_KEYS, aacs->read_data_key, NULL);
+    if (error_code == AACS_SUCCESS) {
+        /* cache read data key */
+        if (!aacs->no_cache) {
+            keycache_save(cache_path, aacs->disc_id, aacs->read_data_key, 16);
+        }
+
+    } else {
+        /* get cached read data key */
+        if (!aacs->no_cache && keycache_find(cache_path, aacs->disc_id, aacs->read_data_key, 16)) {
+            BD_DEBUG(DBG_AACS, "Using cached READ DATA KEY\n");
+            error_code = AACS_SUCCESS;
+        }
+    }
+
+    X_FREE(cache_path);
+
     if (error_code != AACS_SUCCESS) {
         BD_DEBUG(DBG_AACS, "Error reading data keys!\n");
     }
+
     return error_code;
 }
 
@@ -690,7 +806,7 @@ static int _read_pmsn(AACS *aacs, cert_list *hcl)
 
 static int _calc_vuk(AACS *aacs, uint8_t *mk, uint8_t *vuk, config_file *cf)
 {
-    int error_code;
+    int error_code, crypto_err;
 
     /* Skip if retrieved from config file */
     if (memcmp(vuk, empty_key, 16)) {
@@ -722,7 +838,11 @@ static int _calc_vuk(AACS *aacs, uint8_t *mk, uint8_t *vuk, config_file *cf)
 
     /* calculate VUK */
 
-    crypto_aes128d(mk, aacs->vid, vuk);
+    crypto_err = crypto_aes128d(mk, aacs->vid, vuk);
+    if (crypto_err) {
+        LOG_CRYPTO_ERROR(DBG_AACS, "decrypting VUK failed", crypto_err);
+        return AACS_ERROR_UNKNOWN;
+    }
 
     int a;
     for (a = 0; a < 16; a++) {
@@ -743,9 +863,10 @@ static int _calc_vuk(AACS *aacs, uint8_t *mk, uint8_t *vuk, config_file *cf)
 }
 
 /* Function that collects keys from keydb config entry */
-static void _find_config_entry(AACS *aacs, title_entry_list *ce,
+static void _find_config_entry(AACS *aacs, config_file *cf,
                                uint8_t *mk, uint8_t *vuk)
 {
+    title_entry_list *ce = cf->list;
     char str[48];
     char str2[48];
 
@@ -762,16 +883,16 @@ static void _find_config_entry(AACS *aacs, title_entry_list *ce,
         return;
     }
 
-    if (ce->entry.mek) {
-        hexstring_to_hex_array(mk, 16, ce->entry.mek);
+    if (memcmp(ce->entry.mk, empty_key, 16)) {
+        memcpy(mk, ce->entry.mk, 16);
+        memcpy(aacs->mk, mk, 16);
 
         BD_DEBUG(DBG_AACS, "Found media key for %s: %s\n",
                  str_print_hex(str2, ce->entry.discid, 20), str_print_hex(str, mk, 16));
     }
 
-    if (ce->entry.vid) {
-        hexstring_to_hex_array(aacs->vid, sizeof(aacs->vid),
-                               ce->entry.vid);
+    if (memcmp(ce->entry.vid, empty_key, 16)) {
+        memcpy(aacs->vid, ce->entry.vid, 16);
 
         BD_DEBUG(DBG_AACS, "Found volume id for %s: %s\n",
                  str_print_hex(str2, ce->entry.discid, 20), str_print_hex(str, aacs->vid, 16));
@@ -790,20 +911,29 @@ static void _find_config_entry(AACS *aacs, title_entry_list *ce,
         /* count keys */
         unsigned num_uks = 0;
         digit_key_pair_list *ukcursor = ce->entry.uk;
-        while (ukcursor && ukcursor->key_pair.key) {
+        while (ukcursor) {
             num_uks++;
             ukcursor = ukcursor->next;
         }
 
-        /* check against Unit_Key_RO.inf */
+        /* check against Unit_Key_RO.inf, only discard incorrect amount of UKs if VUK or MK+VID is available */
         if (num_uks != aacs->uk->num_uk) {
-            BD_DEBUG(DBG_CRIT | DBG_AACS, "Ignoring unit keys from config file (expected %u keys, found %u)\n",
+            if (_calc_vuk(aacs, mk, vuk, cf) == AACS_SUCCESS) {
+                BD_DEBUG(DBG_CRIT | DBG_AACS, "Ignoring unit keys from config file (expected %u keys, found %u)\n",
+                         aacs->uk->num_uk, num_uks);
+                return;
+            }
+            BD_DEBUG(DBG_CRIT | DBG_AACS, "Incomplete unit keys in config file (expected %u keys, found %u)\n",
                      aacs->uk->num_uk, num_uks);
-            return;
         }
 
         /* get keys */
 
+        if (num_uks > aacs->uk->num_uk) {
+            /* config file has more unit keys than the disc has */
+            aacs->uk->num_uk = num_uks;
+        }
+
         aacs->uk->uk = calloc(aacs->uk->num_uk, sizeof(aacs->uk->uk[0]));
         if (!aacs->uk->uk) {
             return;
@@ -811,8 +941,8 @@ static void _find_config_entry(AACS *aacs, title_entry_list *ce,
 
         num_uks = 0;
         ukcursor = ce->entry.uk;
-        while (ukcursor && ukcursor->key_pair.key) {
-            hexstring_to_hex_array(aacs->uk->uk[num_uks].key, 16, ukcursor->key_pair.key);
+        while (ukcursor) {
+            memcpy(aacs->uk->uk[num_uks].key, ukcursor->key_pair.key, 16);
 
             BD_DEBUG(DBG_AACS, "Unit key %u from keydb entry: %s\n",
                      num_uks,
@@ -897,24 +1027,28 @@ static int _calc_uks(AACS *aacs, config_file *cf)
 
     uint8_t mk[16] = {0}, vuk[16] = {0};
 
-    if (cf) {
-        BD_DEBUG(DBG_AACS, "Searching for keydb config entry...\n");
-        _find_config_entry(aacs, cf->list, mk, vuk);
-
-        /* Skip if retrieved from config file */
-        if (aacs->uk->uk) {
-            return AACS_SUCCESS;
-        }
-    }
-
     if (aacs->uk->num_uk < 1) {
         /* no keys ... */
         return AACS_SUCCESS;
     }
 
-    /* Try to calculate VUK */
+    /* first try cached vuk */
+    vuk_error_code = _calc_vuk(aacs, mk, vuk, NULL);
+    if (vuk_error_code != AACS_SUCCESS) {
 
-    vuk_error_code = _calc_vuk(aacs, mk, vuk, cf);
+        if (cf) {
+            BD_DEBUG(DBG_AACS, "Searching for keydb config entry...\n");
+            _find_config_entry(aacs, cf, mk, vuk);
+
+            /* Skip if retrieved from config file */
+            if (aacs->uk->uk) {
+                return AACS_SUCCESS;
+            }
+        }
+
+        /* Try to calculate VUK */
+        vuk_error_code = _calc_vuk(aacs, mk, vuk, cf);
+    }
 
     BD_DEBUG(DBG_AACS, "Calculate CPS unit keys...\n");
 
@@ -927,6 +1061,7 @@ static int _calc_uks(AACS *aacs, config_file *cf)
     /* decrypt unit keys */
 
     for (i = 0; i < aacs->uk->num_uk; i++) {
+        int crypto_err;
 
         /* error out if VUK calculation fails and encrypted CPS unit is found */
         if (vuk_error_code != AACS_SUCCESS) {
@@ -937,7 +1072,11 @@ static int _calc_uks(AACS *aacs, config_file *cf)
             BD_DEBUG(DBG_AACS | DBG_CRIT, "WARNING: VUK calculation failed but disc seems to be unencrypted.\n");
         }
 
-        crypto_aes128d(vuk, aacs->uk->enc_uk[i].key, aacs->uk->uk[i].key);
+        crypto_err = crypto_aes128d(vuk, aacs->uk->enc_uk[i].key, aacs->uk->uk[i].key);
+        if (crypto_err) {
+            LOG_CRYPTO_ERROR(DBG_AACS, "decrypting unit key failed", crypto_err);
+            return AACS_ERROR_UNKNOWN;
+        }
 
         char str[40];
         BD_DEBUG(DBG_AACS, "Unit key %d: %s\n", i,
@@ -957,10 +1096,11 @@ static int _calc_title_hash(AACS *aacs)
     size_t   size;
     char     str[48];
     int      result = AACS_SUCCESS;
+    int      aacs2 = aacs->cc ? aacs->cc->aacs2 : 0;
 
     size = _read_file(aacs, "AACS" DIR_SEP "Unit_Key_RO.inf", &data);
     if (size > 2048) {
-        aacs->uk = uk_parse(data, size);
+        aacs->uk = uk_parse(data, size, aacs2);
     }
 
     /* failed, try backup */
@@ -968,7 +1108,7 @@ static int _calc_title_hash(AACS *aacs)
         X_FREE(data);
         size = _read_file(aacs, "AACS" DIR_SEP "DUPLICATE" DIR_SEP "Unit_Key_RO.inf", &data);
         if (size > 2048) {
-            aacs->uk = uk_parse(data, size);
+            aacs->uk = uk_parse(data, size, aacs2);
         }
     }
 
@@ -1022,7 +1162,7 @@ static int _get_bus_encryption_enabled(AACS *aacs)
     return 0;
 }
 
-static int _get_bus_encryption_capable(const char *path)
+static int _get_bus_encryption_capable(AACS *aacs, const char *path)
 {
     MMC* mmc = NULL;
     uint8_t drive_cert[92];
@@ -1033,9 +1173,11 @@ static int _get_bus_encryption_capable(const char *path)
     }
 
     if (mmc_read_drive_cert(mmc, drive_cert) == MMC_SUCCESS) {
+        crypto_aacs_title_hash(drive_cert, 92, aacs->drive_cert_hash);
         bec = drive_cert[1] & 1;
         BD_DEBUG(DBG_AACS, "Bus Encryption Capable flag in drive certificate: %d\n", bec);
     } else {
+        memset(aacs->drive_cert_hash, 0, sizeof(aacs->drive_cert_hash));
         BD_DEBUG(DBG_AACS | DBG_CRIT, "Unable to read drive certificate\n");
     }
 
@@ -1065,32 +1207,30 @@ static int _decrypt_unit(AACS *aacs, uint8_t *out_buf, const uint8_t *in_buf, ui
 {
     /* inbuf == NULL means in-place decryption */
 
-    gcry_cipher_hd_t gcry_h;
-    int a;
+    int a, crypto_err;
     uint8_t key[16];
 
     if (BD_UNLIKELY(in_buf != NULL)) {
         memcpy(out_buf, in_buf, 16); /* first 16 bytes are plain */
     }
 
-    gcry_cipher_open(&gcry_h, GCRY_CIPHER_AES, GCRY_CIPHER_MODE_ECB, 0);
-    gcry_cipher_setkey(gcry_h, aacs->uk->uk[curr_uk].key, 16);
-    gcry_cipher_encrypt(gcry_h, key, 16, out_buf, 16); /* here out_buf is plain data fron in_buf */
-    gcry_cipher_close(gcry_h);
+    crypto_err = crypto_aes128e(aacs->uk->uk[curr_uk].key, out_buf, key);
+    if (crypto_err) {
+        LOG_CRYPTO_ERROR(DBG_AACS, "unit key derivation failed", crypto_err);
+    }
 
     for (a = 0; a < 16; a++) {
-        key[a] ^= out_buf[a]; /* here out_buf is plain data fron in_buf */
+        key[a] ^= out_buf[a]; /* here out_buf is plain data from in_buf */
     }
 
-    gcry_cipher_open(&gcry_h, GCRY_CIPHER_AES, GCRY_CIPHER_MODE_CBC, 0);
-    gcry_cipher_setkey(gcry_h, key, 16);
-    gcry_cipher_setiv(gcry_h, aacs_iv, 16);
     if (BD_UNLIKELY(in_buf != NULL)) {
-        gcry_cipher_decrypt(gcry_h, out_buf + 16, ALIGNED_UNIT_LEN - 16, in_buf + 16, ALIGNED_UNIT_LEN - 16);
+        crypto_err = crypto_aacs_decrypt(key, out_buf + 16, ALIGNED_UNIT_LEN - 16, in_buf + 16, ALIGNED_UNIT_LEN - 16);
     } else {
-        gcry_cipher_decrypt(gcry_h, out_buf + 16, ALIGNED_UNIT_LEN - 16, NULL, 0);
+        crypto_err = crypto_aacs_decrypt(key, out_buf + 16, ALIGNED_UNIT_LEN - 16, NULL, 0);
+    }
+    if (crypto_err) {
+        LOG_CRYPTO_ERROR(DBG_AACS, "decrypting unit failed", crypto_err);
     }
-    gcry_cipher_close(gcry_h);
 
     if (_verify_ts(out_buf)) {
         return 1;
@@ -1099,17 +1239,6 @@ static int _decrypt_unit(AACS *aacs, uint8_t *out_buf, const uint8_t *in_buf, ui
     return 0;
 }
 
-static void _decrypt_bus(AACS *aacs, uint8_t *buf)
-{
-    gcry_cipher_hd_t gcry_h;
-
-    gcry_cipher_open(&gcry_h, GCRY_CIPHER_AES, GCRY_CIPHER_MODE_CBC, 0);
-    gcry_cipher_setkey(gcry_h, aacs->read_data_key, 16);
-    gcry_cipher_setiv(gcry_h, aacs_iv, 16);
-    gcry_cipher_decrypt(gcry_h, buf + 16, SECTOR_LEN - 16, NULL, 0);
-    gcry_cipher_close(gcry_h);
-}
-
 /*
  * libaacs API
  */
@@ -1133,6 +1262,7 @@ const char *aacs_error_str(int err)
        [-AACS_ERROR_MMC_OPEN]         = "Failed opening MMC device",
        [-AACS_ERROR_MMC_FAILURE]      = "MMC failure",
        [-AACS_ERROR_NO_DK]            = "No matching device key",
+       [-AACS_ERROR_UNKNOWN]          = "Error",
     };
     err = -err;
     if (err < 0 || (size_t)err >= sizeof(str) / sizeof(str[0]) || !str[err]) {
@@ -1141,7 +1271,7 @@ const char *aacs_error_str(int err)
     return str[err];
 }
 
-/* aacs_open2() wrapper for backwards compability */
+/* aacs_open2() wrapper for backwards compatibility */
 AACS *aacs_open(const char *path, const char *configfile_path)
 {
     int error_code;
@@ -1156,7 +1286,7 @@ AACS *aacs_open(const char *path, const char *configfile_path)
     return NULL;
 }
 
-/* aacs_open_device() wrapper for backward compability */
+/* aacs_open_device() wrapper for backward compatibility */
 AACS *aacs_open2(const char *path, const char *configfile_path, int *error_code)
 {
     AACS *aacs = aacs_init();
@@ -1206,12 +1336,16 @@ int aacs_open_device(AACS *aacs, const char *path, const char *configfile_path)
 
     aacs->path = path ? str_dup(path) : NULL;
 
+    aacs->cc = _read_cc_any(aacs);
+    aacs->bee = _get_bus_encryption_enabled(aacs);
+    aacs->bec = -1;
+
     error_code = _calc_title_hash(aacs);
     if (error_code != AACS_SUCCESS) {
         return error_code;
     }
 
-    cf = keydbcfg_config_load(configfile_path);
+    cf = keydbcfg_config_load(configfile_path, aacs->disc_id);
 
     BD_DEBUG(DBG_AACS, "Starting AACS waterfall...\n");
     error_code = _calc_uks(aacs, cf);
@@ -1219,20 +1353,20 @@ int aacs_open_device(AACS *aacs, const char *path, const char *configfile_path)
         BD_DEBUG(DBG_AACS, "Failed to initialize AACS!\n");
     }
 
-    aacs->cc = _read_cc_any(aacs);
-
-    aacs->bee = _get_bus_encryption_enabled(aacs);
-    aacs->bec = _get_bus_encryption_capable(path);
-
-    if (error_code == AACS_SUCCESS && aacs->bee && aacs->bec) {
+    if (error_code == AACS_SUCCESS && aacs->bee) {
 
         if (!cf) {
             return AACS_ERROR_NO_CONFIG;
         }
 
+        if (aacs->bec < 0) {
+            aacs->bec = _get_bus_encryption_capable(aacs, path);
+        }
+        if (aacs->bec > 0) {
         error_code = _read_read_data_key(aacs, cf->host_cert_list);
         if (error_code != AACS_SUCCESS) {
             BD_DEBUG(DBG_AACS | DBG_CRIT, "Unable to initialize bus encryption required by drive and disc\n");
+            }
         }
     }
 
@@ -1264,10 +1398,15 @@ void aacs_close(AACS *aacs)
 
 static void _decrypt_unit_bus(AACS *aacs, uint8_t *buf)
 {
-    if (aacs->bee && aacs->bec) {
+    if (aacs->bee && aacs->bec > 0) {
         unsigned int i;
+        int crypto_err;
         for (i = 0; i < ALIGNED_UNIT_LEN; i += SECTOR_LEN) {
-            _decrypt_bus(aacs, buf + i);
+            //_decrypt_bus(aacs, buf + i);
+            crypto_err = crypto_aacs_decrypt(aacs->read_data_key, buf + i + 16, SECTOR_LEN - 16, NULL, 0);
+            if (crypto_err) {
+                LOG_CRYPTO_ERROR(DBG_AACS, "bus decrypting failed", crypto_err);
+            }
         }
     }
 }
@@ -1367,7 +1506,7 @@ const uint8_t *aacs_get_bdj_root_cert_hash(AACS *aacs)
 const uint8_t *aacs_get_mk(AACS *aacs)
 {
     if (!memcmp(aacs->mk, empty_key, sizeof(aacs->mk))) {
-        config_file *cf = keydbcfg_config_load(NULL);
+        config_file *cf = keydbcfg_config_load(NULL, NULL);
         if (cf) {
             uint8_t mk[16] = {0};
             if (_calc_mk(aacs, mk, cf->pkl, cf->dkl) == AACS_SUCCESS) {
@@ -1395,7 +1534,7 @@ const uint8_t *aacs_get_vid(AACS *aacs)
             return aacs->vid;
         }
 
-        config_file *cf = keydbcfg_config_load(NULL);
+        config_file *cf = keydbcfg_config_load(NULL, NULL);
         if (cf) {
             _read_vid(aacs, cf->host_cert_list);
 
@@ -1414,7 +1553,7 @@ const uint8_t *aacs_get_vid(AACS *aacs)
 const uint8_t *aacs_get_pmsn(AACS *aacs)
 {
     if (!memcmp(aacs->pmsn, empty_key, sizeof(aacs->pmsn))) {
-        config_file *cf = keydbcfg_config_load(NULL);
+        config_file *cf = keydbcfg_config_load(NULL, NULL);
         if (cf) {
             _read_pmsn(aacs, cf->host_cert_list);
 
@@ -1521,6 +1660,9 @@ void aacs_free_rl(AACS_RL_ENTRY **rl)
 
 uint32_t aacs_get_bus_encryption(AACS *aacs)
 {
+    if (aacs->bec < 0) {
+        aacs->bec = _get_bus_encryption_capable(aacs, aacs->path);
+    }
   return (aacs->bee * AACS_BUS_ENCRYPTION_ENABLED) |
          (aacs->bec * AACS_BUS_ENCRYPTION_CAPABLE);
 }
