@@ -41,14 +41,20 @@ static const uint8_t *_record(MKB *mkb, uint8_t type, size_t *rec_len)
     while (pos + 4 <= mkb->size) {
         len = MKINT_BE24(mkb->buf + pos + 1);
 
-        if (rec_len) {
-            *rec_len = len;
-        }
-
         if (mkb->buf[pos] == type) {
             BD_DEBUG(DBG_MKB, "Retrieved MKB record 0x%02x (%p)\n", type,
                   (void*)(mkb->buf + pos));
 
+            if (len > mkb->size - pos) {
+                BD_DEBUG(DBG_MKB | DBG_CRIT, "Ignoring truncated MKB record 0x%02x @ %zu, size %zu (%p)\n", type, pos, len,
+                         (void*)(mkb->buf + pos));
+                return NULL;
+            }
+
+            if (rec_len) {
+                *rec_len = len;
+            }
+
             return mkb->buf + pos;
         }
 
@@ -77,6 +83,7 @@ MKB *mkb_init(uint8_t *data, size_t len)
     mkb->buf  = data;
 
     BD_DEBUG(DBG_MKB, "MKB size: %u\n", (unsigned)mkb->size);
+    BD_DEBUG(DBG_MKB, "MKB type: 0x%x\n", mkb_type(mkb));
     BD_DEBUG(DBG_MKB, "MKB version: %d\n", mkb_version(mkb));
 
     return mkb;
@@ -106,11 +113,32 @@ size_t mkb_data_size(MKB *mkb)
         pos += MKINT_BE24(mkb->buf + pos + 1);
     }
 
+    if (pos > mkb->size) {
+        BD_DEBUG(DBG_MKB | DBG_CRIT, "mkb_data_size(): invalid or truncated MKB\n");
+        return mkb->size;
+    }
+
+    BD_DEBUG(DBG_MKB, "MKB data size %zu bytes\n", pos);
     return pos;
 }
 
+static const uint8_t *_simple_record(MKB *mkb, uint8_t type, size_t *len)
+{
+    const uint8_t *rec = _record(mkb, type, len);
 
-uint8_t mkb_type(MKB *mkb)
+    if (*len < 4) {
+        return NULL;
+    }
+    if (rec) {
+        rec += 4;
+        *len -= 4;
+    }
+
+    return rec;
+}
+
+
+uint32_t mkb_type(MKB *mkb)
 {
     const uint8_t *rec = mkb_type_and_version_record(mkb);
 
@@ -144,77 +172,49 @@ const uint8_t *mkb_type_and_version_record(MKB *mkb)
     return rec;
 }
 
-
 const uint8_t *mkb_host_revokation_entries(MKB *mkb, size_t *len)
 {
-    const uint8_t *rec = _record(mkb, 0x21, len);
-
-    if (*len < 4) {
-        return NULL;
-    }
-    if (rec) {
-        rec += 4;
-        *len -= 4;
-    }
-
-    return rec;
+    return _simple_record(mkb, 0x21, len);
 }
 
 const uint8_t *mkb_drive_revokation_entries(MKB *mkb, size_t *len)
 {
-    const uint8_t *rec = _record(mkb, 0x20, len);
-
-    if (*len < 4) {
-        return NULL;
-    }
-    if (rec) {
-        rec += 4;
-        *len -= 4;
-    }
-
-    return rec;
+    return _simple_record(mkb, 0x20, len);
 }
 
 const uint8_t *mkb_subdiff_records(MKB *mkb, size_t *len)
 {
-    const uint8_t *rec = _record(mkb, 0x04, len);
-
-    if (*len < 4) {
-        return NULL;
-    }
-    if (rec) {
-        rec += 4;
-        *len -= 4;
-    }
-
-    return rec;
+    return _simple_record(mkb, 0x04, len);
 }
 
 const uint8_t *mkb_cvalues(MKB *mkb, size_t *len)
 {
-    const uint8_t *rec = _record(mkb, 0x05, len);
-
-    if (*len < 4) {
-        return NULL;
-    }
-    if (rec) {
-        rec += 4;
-        *len -= 4;
-    }
-
-    return rec;
+    return _simple_record(mkb, 0x05, len);
 }
 
 const uint8_t *mkb_mk_dv(MKB *mkb)
 {
+    const uint8_t *rec;
+    uint8_t dv_record;
     size_t len;
-    const uint8_t *rec = _record(mkb, 0x81, &len);
 
-    if (len < 20) {
-        return NULL;
+    switch (mkb_type(mkb)) {
+        case MKB_20_CATEGORY_C:
+        case MKB_21_CATEGORY_C:
+            dv_record = 0x86;
+            break;
+        case MKB_TYPE_3:
+        case MKB_TYPE_4:
+        case MKB_TYPE_10_CLASS_II:
+        default:
+            dv_record = 0x81;
+            break;
     }
-    if (rec) {
-        rec += 4;
+
+    rec = _simple_record(mkb, dv_record, &len);
+
+    if (len < 16) {
+        return NULL;
     }
 
     return rec;
@@ -222,18 +222,7 @@ const uint8_t *mkb_mk_dv(MKB *mkb)
 
 const uint8_t *mkb_signature(MKB *mkb, size_t *len)
 {
-    const uint8_t *rec = _record(mkb, 0x02, len);
-
-    if (*len < 4) {
-        return NULL;
-    }
-    if (rec) {
-        rec += 4;
-        *len -= 4;
-    }
-
-    return rec;
-
+    return _simple_record(mkb, 0x02, len);
 }
 
 static int _cert_is_revoked(const uint8_t *rl, size_t rl_size, const uint8_t *cert_id_bin)
